! WHIZARD 1.12 Jun 20 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module jetset_interface

  use kinds, only: default !NODEP!
  use kinds, only: double !NODEP!
  use diagnostics, only: msg_buffer, msg_message, msg_fatal
  use limits, only: PYTHIA_EXTERNAL_PROCESS, MB_PER_FB_EXPONENT

  implicit none
  private

  integer, parameter, public :: PYGIVE_LEN = 100

  public :: jetset_init, pythia_init
  public :: jetset_fragment
  public :: pythia_fragment
  public :: pythia_up_fill
  public :: jetset_write
  public :: jetset_suppress_banner

  integer :: pythia_isub
  real(kind=double) :: pythia_sqrts, pythia_integral
  integer :: pythia_nup
  integer, dimension(20,7) :: pythia_kup
  real(kind=double), dimension(20,5) :: pythia_pup
!  integer :: pythia_nfup
!  integer, dimension(10,2) :: pythia_ifup
!  real(kind=double) :: pythia_q2up
  save

contains

  subroutine jetset_init (chin)
    character(len=PYGIVE_LEN), intent(in) :: chin
    if (chin /= ' ') call pygive (chin)
  end subroutine jetset_init

  subroutine pythia_init (n_external, process_id, beam, target, &
       &                  fixed_energy, sqrts, integral, chin)
    integer, intent(in) :: n_external
    character(len=*), intent(in) :: process_id, beam, target
    logical, intent(in) :: fixed_energy
    real(kind=default), intent(in) :: sqrts, integral
    character(len=PYGIVE_LEN), intent(in) :: chin
    character(len=3) :: prc_str
    if (n_external > 20) then
       write (msg_buffer, "(A,1x,I5)") &
            & " Number of external particles:", n_external
       call msg_message
       call msg_fatal &
            & (" PYTHIA fragmentation is possible for max. 20 particles.")
    end if
    pythia_nup = n_external
    pythia_isub = PYTHIA_EXTERNAL_PROCESS
    pythia_sqrts = sqrts
    pythia_integral = integral / 10.d0**MB_PER_FB_EXPONENT
    call pyupin (pythia_isub, trim(process_id), pythia_integral)
    if (chin /= ' ') call pygive (chin)           ! User intervention
    call pygive ("MSTP(61)=0;  MSTP(71)=0")       ! switch off ISR and FSR
    if (fixed_energy) then
       call pygive ("MSTP(11)=0")                 ! No electron ISR
    else
       pythia_sqrts = pythia_sqrts &
            &         * 1.00000000001_double      ! suppress x=1 warnings
    end if
    write (prc_str, "(I3)")  pythia_isub
    call pygive ("MSEL=0; MSUB("//prc_str//")=1") ! select user process
    call pyinit ("CMS", beam, target, pythia_sqrts)
  end subroutine pythia_init

  subroutine jetset_fragment (color_flow, anticolor_flow)
    integer, dimension(:), intent(in) :: color_flow, anticolor_flow
    integer, dimension(size(color_flow)) :: cflow, aflow
    integer :: i
    call pyhepc (2)
    cflow = color_flow
    aflow = anticolor_flow
    do i=1, size(cflow)
       if (cflow(i)/=0 .and. aflow(i)==0) then
          call make_color_string (i, cflow, aflow)
       end if
    end do
    do i=1, size(cflow)
       if (aflow(i)/=0 .and. cflow(i)==0) then
          call make_color_string (i, aflow, cflow)
       end if
    end do
    do i=1, size(cflow)
       if (cflow(i)/=0 .and. aflow(i)/=0) then
          call make_color_string (i, cflow, aflow)
       end if
    end do
    call pyexec ()
    call pyhepc (1)
  contains
  subroutine make_color_string (i0, flow1, flow2)
    integer, intent(in) :: i0
    integer, dimension(:), intent(inout) :: flow1, flow2
    integer, dimension(size(flow1)) :: ijoin
    integer :: njoin
    integer :: i1, i2, j
    i1 = i0
    ijoin = 0
    njoin = 0
    do j=1, size(ijoin)
       if (i1 <= 2) exit
       ijoin(j) = i1 - 2
       njoin = njoin + 1
       i2 = i1
       i1 = flow1(i2)
       flow1(i2) = 0
       flow2(i2) = 0
    end do
    if (njoin > 1)  call pyjoin (njoin, ijoin)
  end subroutine make_color_string
  end subroutine jetset_fragment

  subroutine pythia_fragment (m, p, code, color_flow, anticolor_flow)
    real(kind=default), dimension(:), intent(in) :: m
    real(kind=default), dimension(:,0:), intent(in) :: p
    integer, dimension(:), intent(in) :: code, color_flow, anticolor_flow
    integer :: n
    n = pythia_nup
    pythia_kup (:n,1) = 1
    pythia_kup (:n,2) = code (:n)
    pythia_kup (:n,3) = 0
    pythia_kup (:2,4)  = 0
    pythia_kup (3:n,4) = color_flow (3:n)
    pythia_kup (1:2,5) = 0
    pythia_kup (3:n,5) = anticolor_flow (3:n)
    pythia_kup (1:2,6) = anticolor_flow (1:2)
    pythia_kup (3:n,6) = 0
    pythia_kup (1:2,7) = color_flow (1:2)
    pythia_kup (3:n,7) = 0
    pythia_pup (1:n,1:3) = p(1:n,1:3)
    pythia_pup (1:n,4)   = p(1:n,0)
    pythia_pup (1:n,5)   = m(1:n)
    call pyevnt ()
    call pyhepc (1)
  end subroutine pythia_fragment

  subroutine pythia_up_fill (integral)
    real(kind=double), intent(out) :: integral
    integer :: nup, kup, nfup, ifup
    real(kind=double) :: pup, q2up
    common /PYUPPR/ nup, kup(20,7), nfup, ifup(10,2), pup(20,5), q2up(0:10)
    save /PYUPPR/
    nup = pythia_nup
    kup = pythia_kup
    pup = pythia_pup
    integral = pythia_integral
!    nfup = pythia_nfup
!    ifup = pythia_ifup
!    q2up = pythia_q2up
  end subroutine pythia_up_fill

  subroutine jetset_write (mlist)
    integer, intent(in) :: mlist
    call pyhepc (2)
    call pylist (mlist)
  end subroutine jetset_write

  subroutine jetset_suppress_banner
    integer, dimension(200) :: mstu, mstj
    real(kind=double), dimension(200) :: paru, parj
    common /PYDAT1/ mstu, paru, mstj, parj
    mstu(12) = 0
  end subroutine jetset_suppress_banner

end module jetset_interface

subroutine pyupev (isub_dummy, sigev)
  use kinds, only: double !NODEP!
  use jetset_interface, only: pythia_up_fill
  implicit none
  integer, intent(in) :: isub_dummy
  real(kind=double), intent(out) :: sigev
  real(kind=double) :: integral
  call pythia_up_fill (integral)
  sigev = integral
end subroutine pyupev

