! WHIZARD 1.12 Jun 20 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module monte_carlo_states

  use kinds, only: default !NODEP!
  use mpi90, only: mpi90_broadcast, mpi90_rank !NODEP!
  use limits, only: PROCESS_ID_LEN, MAX_CHANNELS, WHIZARD_ROOT
  use limits, only: LOGTABLE_LINE_LENGTH
  use limits, only: DEFAULT_FILENAME
  use jetset_interface, only: PYGIVE_LEN
  use file_utils, only: free_unit !NODEP!
  use parameters !NODEP!
  use particles
  use decay_forests
  use events
  use cuts
  use beams
  use diagnostics, only: msg_buffer, msg_level
  use diagnostics, only: msg_result, msg_message, msg_fatal, msg_error
  use process_interface !NODEP!
  use tao_random_numbers !NODEP!
  use vamp_rest, only: vamp_grids !NODEP!
  use vamp_rest, only: vamp_history, vamp_create_history, vamp_write_history !NODEP!

  implicit none
  private

  public :: create, destroy
  public :: read, write
  public :: mpi90_broadcast
  public :: checksum
  public :: checksum_error
  public :: mc_write_header
  public :: mc_write_hline, mc_write_dline
  public :: mc_write_result
  public :: mc_read_result
  public :: mc_get_efficiency
  public :: mc_get_accuracy

  type, public :: monte_carlo_input
     character(len=PROCESS_ID_LEN) :: process_id
     real(kind=default) :: sqrts
     real(kind=default) :: luminosity
     logical :: polarized_beams
     integer, dimension(2,5) :: calls
     integer :: seed
     real(kind=default) :: accuracy_goal
     real(kind=default) :: efficiency_goal
     logical :: stratified
     logical :: use_efficiency
     real(kind=default) :: weights_power
     logical :: write_grids
     logical :: write_all_grids
     logical :: read_grids
     logical :: read_grids_force
     real(kind=default), dimension(MAX_CHANNELS) :: initial_weights
     integer :: apply_mappings
     logical :: generate_phase_space
     logical :: read_phase_space
     logical :: overwrite_phase_space
     logical :: phase_space_only
     integer :: off_shell_lines
     integer :: exchange_lines
     integer :: extra_off_shell_lines
     logical :: s_channel_resonance
     logical :: single_off_shell_decays
     logical :: double_off_shell_decays
     logical :: single_off_shell_branchings
     logical :: double_off_shell_branchings
     logical :: massive_fsr
     real(kind=default) :: default_jet_cut
     real(kind=default) :: default_mass_cut
     real(kind=default) :: default_energy_cut
     real(kind=default) :: default_q_cut
     integer :: n_events
     integer :: n_events_warmup
     logical :: unweighted
     real(kind=default) :: safety_factor
     logical :: write_events_raw
     logical :: write_events
     integer :: write_events_format
     logical :: read_events_raw
     logical :: read_events_force
     logical :: fragment
     integer :: fragmentation_method
     character(len=PYGIVE_LEN) :: pygive
     logical :: show_pythia_banner
     integer :: chattiness
     logical :: screen_results
     logical :: screen_histograms
     logical :: screen_diagnostics
     logical :: write_logfile
     logical :: show_input
     logical :: show_results
     logical :: show_phase_space
     logical :: show_cuts
     logical :: show_histories
     logical :: show_history
     logical :: show_weights
     logical :: show_event
     logical :: show_histograms
     logical :: show_overflow
     logical :: show_excess
     integer :: plot_width, plot_height
     type(parameter_set) :: par
     type(beam_parameter_set), dimension(2) :: beam
  end type monte_carlo_input

  type, public :: monte_carlo_state
     character(len=PROCESS_ID_LEN) :: process_id
     real(kind=default) :: sqrts
     real(kind=default) :: luminosity
     logical :: polarized_beams
     integer, dimension(2,5) :: calls
     integer :: seed
     real(kind=default) :: accuracy_goal
     real(kind=default) :: efficiency_goal
     logical :: stratified
     logical :: use_efficiency
     real(kind=default) :: weights_power
     logical :: write_grids
     logical :: write_all_grids
     logical :: read_grids
     logical :: read_grids_force
     real(kind=default), dimension(MAX_CHANNELS) :: initial_weights
     integer :: apply_mappings
     logical :: generate_phase_space
     logical :: read_phase_space
     logical :: overwrite_phase_space
     logical :: phase_space_only
     integer :: off_shell_lines
     integer :: exchange_lines
     integer :: extra_off_shell_lines
     logical :: s_channel_resonance
     logical :: single_off_shell_decays
     logical :: double_off_shell_decays
     logical :: single_off_shell_branchings
     logical :: double_off_shell_branchings
     logical :: massive_fsr
     real(kind=default) :: default_jet_cut
     real(kind=default) :: default_mass_cut
     real(kind=default) :: default_energy_cut
     real(kind=default) :: default_q_cut
     integer :: n_events
     integer :: n_events_warmup
     logical :: unweighted
     real(kind=default) :: safety_factor
     logical :: write_events_raw
     logical :: write_events
     integer :: write_events_format
     logical :: read_events_raw
     logical :: read_events_force
     logical :: fragment
     integer :: fragmentation_method
     character(len=PYGIVE_LEN) :: pygive
     logical :: show_pythia_banner
     integer :: chattiness
     logical :: screen_results
     logical :: screen_histograms
     logical :: screen_diagnostics
     logical :: write_logfile
     logical :: show_input
     logical :: show_results
     logical :: show_phase_space
     logical :: show_cuts
     logical :: show_histories
     logical :: show_history
     logical :: show_weights
     logical :: show_event
     logical :: show_histograms
     logical :: show_overflow
     logical :: show_excess
     integer :: plot_width, plot_height
     type(parameter_set) :: par
     type(beam_parameter_set), dimension(2) :: beam
     integer :: n_in, n_out, n_tot, n_col, n_hel, n_flv
     integer :: n_col_in, n_col_out, n_hel_in, n_hel_out, n_flv_in, n_flv_out
     integer, dimension(:,:), pointer :: code
     integer, dimension(:,:), pointer :: code_in
     integer, dimension(:,:), pointer :: code_out
     integer, dimension(:,:), pointer :: hel_state
     integer, dimension(:,:), pointer :: hel_state_in
     integer, dimension(:,:), pointer :: hel_state_out
     integer, dimension(:,:), pointer :: col_flow
     integer, dimension(:,:), pointer :: acl_flow
     real(kind=default), dimension(:,:), pointer :: rho_in
     integer :: n_iterations_whizard
     integer :: n_iterations_vamp
     integer :: n_channels
     type(decay_forest) :: f
     type(cut_array) :: cc
     integer :: checksum
     integer :: it, it_vamp
     type(vamp_grids) :: g, g_best
     type(vamp_history), dimension(:), pointer :: history
     type(vamp_history), dimension(:,:), pointer :: histories
     character(len=LOGTABLE_LINE_LENGTH), dimension(:), pointer :: logtable
     real(kind=default), dimension(:,:), pointer :: weights
     logical :: actual_integration_done
     real(kind=default) :: integral, error, chi2, efficiency, accuracy
     type(event) :: evt
     type(tao_random_state) :: rng
  end type monte_carlo_state

  integer, parameter, public :: &
       & NO_FRAGMENTATION=0, JETSET_FRAGMENTATION=1, PYTHIA_FRAGMENTATION=2

  type(monte_carlo_state), public :: mcs

  interface create
     module procedure mc_input_create
  end interface
  interface destroy
     module procedure mc_input_destroy
  end interface
  interface read
     module procedure mc_input_read_unit
     module procedure mc_input_read_name
  end interface
  interface write
     module procedure mc_input_write_unit
     module procedure mc_input_write_name
  end interface
  interface mpi90_broadcast
     module procedure mc_input_broadcast
  end interface
  interface mpi90_broadcast
     module procedure character_broadcast
  end interface
  
  interface mpi90_broadcast
     module procedure parameter_set_broadcast
  end interface
  interface create
     module procedure mc_state_create
  end interface
  interface destroy
     module procedure mc_state_destroy
  end interface
  interface write
     module procedure mc_state_write_unit
     module procedure mc_state_write_name
  end interface
  interface checksum
     module procedure mc_checksum
  end interface
  interface checksum_error
     module procedure mc_checksum_error
  end interface
  interface mc_write_header
     module procedure mc_write_header_unit
  end interface
  interface mc_write_hline
     module procedure mc_write_hline_unit
  end interface
  interface mc_write_dline
     module procedure mc_write_dline_unit
  end interface
  interface mc_write_result
     module procedure mc_write_result_string
     module procedure mc_write_result_unit
     module procedure mc_write_logtable_unit
  end interface
  interface mc_read_result
     module procedure mc_read_logtable_unit
  end interface

contains

  subroutine mc_input_create(mci)
    type(monte_carlo_input), intent(out) :: mci
    call create(mci%beam)
    mci%process_id = "[undefined]"
    mci%sqrts = 1000
    mci%luminosity = 0
    mci%polarized_beams = .false.
    mci%calls = 0
    call system_clock(mci%seed)
    mci%accuracy_goal = 0
    mci%efficiency_goal = 100
    mci%stratified = .true.
    mci%use_efficiency = .false.
    mci%weights_power = 0.25_default
    mci%write_grids = .true.
    mci%write_all_grids = .false.
    mci%read_grids = .false.
    mci%read_grids_force = .false.
    mci%initial_weights = 1
    mci%apply_mappings = 0
    mci%generate_phase_space = .true.
    mci%read_phase_space = .true.
    mci%overwrite_phase_space = .true.
    mci%phase_space_only = .false.
    mci%off_shell_lines = 2
    mci%exchange_lines = 1
    mci%extra_off_shell_lines = 0
    mci%s_channel_resonance = .false.
    mci%single_off_shell_decays = .true.
    mci%double_off_shell_decays = .false.
    mci%single_off_shell_branchings = .true.
    mci%double_off_shell_branchings = .true.
    mci%massive_fsr = .true.
    mci%default_jet_cut = 10
    mci%default_mass_cut = 10
    mci%default_energy_cut = 10
    mci%default_q_cut = 10
    mci%n_events = 0
    mci%n_events_warmup = 0
    mci%unweighted = .true.
    mci%safety_factor = 1
    mci%write_events_raw = .true.
    mci%write_events = .false.
    mci%write_events_format = HEPEVT_FMT
    mci%read_events_raw = .false.
    mci%read_events_force = .false.
    mci%fragment = .false.
    mci%fragmentation_method = JETSET_FRAGMENTATION
    mci%pygive = ' '
    mci%chattiness = 4
    mci%screen_results = .true.
    mci%screen_histograms = .false.
    mci%screen_diagnostics = .false.
    mci%show_pythia_banner = .true.
    mci%write_logfile = .true.
    mci%show_input = .true.
    mci%show_results = .true.
    mci%show_phase_space = .false.
    mci%show_cuts = .true.
    mci%show_histories = .true.
    mci%show_history = .true.
    mci%show_weights = .true.
    mci%show_event = .false.
    mci%show_histograms = .false.
    mci%show_overflow = .false.
    mci%show_excess = .true.
    mci%plot_width = 130
    mci%plot_height = 90
    call init (mci%par)
  end subroutine mc_input_create

  subroutine mc_input_destroy(mci)
    type(monte_carlo_input), intent(inout) :: mci
    call destroy(mci%beam)
  end subroutine mc_input_destroy

  subroutine mc_input_read_unit(u, mci)
    integer, intent(in) :: u
    type(monte_carlo_input), intent(inout) :: mci
    character(len=PROCESS_ID_LEN) :: process_id
    real(kind=default) :: sqrts
    real(kind=default) :: luminosity
    logical :: polarized_beams
    integer, dimension(2,5) :: calls
    integer :: seed
    real(kind=default) :: accuracy_goal
    real(kind=default) :: efficiency_goal
    logical :: stratified
    logical :: use_efficiency
    real(kind=default) :: weights_power
    logical :: write_grids
    logical :: write_all_grids
    logical :: read_grids
    logical :: read_grids_force
    real(kind=default), dimension(MAX_CHANNELS) :: initial_weights
    integer :: apply_mappings
    logical :: generate_phase_space
    logical :: read_phase_space
    logical :: overwrite_phase_space
    logical :: phase_space_only
    integer :: off_shell_lines
    integer :: exchange_lines
    integer :: extra_off_shell_lines
    logical :: s_channel_resonance
    logical :: single_off_shell_decays
    logical :: double_off_shell_decays
    logical :: single_off_shell_branchings
    logical :: double_off_shell_branchings
    logical :: massive_fsr
    real(kind=default) :: default_jet_cut
    real(kind=default) :: default_mass_cut
    real(kind=default) :: default_energy_cut
    real(kind=default) :: default_q_cut
    integer :: n_events
    integer :: n_events_warmup
    logical :: unweighted
    real(kind=default) :: safety_factor
    logical :: write_events_raw
    logical :: write_events
    integer :: write_events_format
    logical :: read_events_raw
    logical :: read_events_force
    logical :: fragment
    integer :: fragmentation_method
    character(len=PYGIVE_LEN) :: pygive
    logical :: show_pythia_banner
    integer :: chattiness
    logical :: screen_results
    logical :: screen_histograms
    logical :: screen_diagnostics
    logical :: write_logfile
    logical :: show_input
    logical :: show_results
    logical :: show_phase_space
    logical :: show_cuts
    logical :: show_histories
    logical :: show_history
    logical :: show_weights
    logical :: show_event
    logical :: show_histograms
    logical :: show_overflow
    logical :: show_excess
    integer :: plot_width, plot_height
    namelist /process_input/ process_id
    namelist /process_input/ sqrts
    namelist /process_input/ luminosity
    namelist /process_input/ polarized_beams
    namelist /integration_input/ calls
    namelist /integration_input/ seed
    namelist /integration_input/ accuracy_goal
    namelist /integration_input/ efficiency_goal
    namelist /integration_input/ stratified
    namelist /integration_input/ use_efficiency
    namelist /integration_input/ weights_power
    namelist /integration_input/ write_grids
    namelist /integration_input/ write_all_grids
    namelist /integration_input/ read_grids
    namelist /integration_input/ read_grids_force
    namelist /integration_input/ initial_weights
    namelist /integration_input/ apply_mappings
    namelist /integration_input/ generate_phase_space
    namelist /integration_input/ read_phase_space
    namelist /integration_input/ overwrite_phase_space
    namelist /integration_input/ phase_space_only
    namelist /integration_input/ off_shell_lines
    namelist /integration_input/ exchange_lines
    namelist /integration_input/ extra_off_shell_lines
    namelist /integration_input/ s_channel_resonance
    namelist /integration_input/ single_off_shell_decays
    namelist /integration_input/ double_off_shell_decays
    namelist /integration_input/ single_off_shell_branchings
    namelist /integration_input/ double_off_shell_branchings
    namelist /integration_input/ massive_fsr
    namelist /integration_input/ default_jet_cut
    namelist /integration_input/ default_mass_cut
    namelist /integration_input/ default_energy_cut
    namelist /integration_input/ default_q_cut
    namelist /simulation_input/ n_events
    namelist /simulation_input/ n_events_warmup
    namelist /simulation_input/ unweighted
    namelist /simulation_input/ safety_factor
    namelist /simulation_input/ write_events_raw
    namelist /simulation_input/ write_events
    namelist /simulation_input/ write_events_format
    namelist /simulation_input/ read_events_raw
    namelist /simulation_input/ read_events_force
    namelist /simulation_input/ fragment
    namelist /simulation_input/ fragmentation_method
    namelist /simulation_input/ pygive
    namelist /diagnostics_input/ chattiness
    namelist /diagnostics_input/ screen_results
    namelist /diagnostics_input/ screen_histograms
    namelist /diagnostics_input/ screen_diagnostics
    namelist /diagnostics_input/ show_pythia_banner
    namelist /diagnostics_input/ write_logfile
    namelist /diagnostics_input/ show_input
    namelist /diagnostics_input/ show_results
    namelist /diagnostics_input/ show_phase_space
    namelist /diagnostics_input/ show_cuts
    namelist /diagnostics_input/ show_histories
    namelist /diagnostics_input/ show_history
    namelist /diagnostics_input/ show_weights
    namelist /diagnostics_input/ show_event
    namelist /diagnostics_input/ show_histograms
    namelist /diagnostics_input/ show_overflow
    namelist /diagnostics_input/ show_excess
    namelist /diagnostics_input/ plot_width, plot_height
    process_id = mci%process_id
    sqrts = mci%sqrts
    luminosity = mci%luminosity
    polarized_beams = mci%polarized_beams
    calls = mci%calls
    seed = mci%seed
    accuracy_goal = mci%accuracy_goal
    efficiency_goal = mci%efficiency_goal
    stratified = mci%stratified
    use_efficiency = mci%use_efficiency
    weights_power = mci%weights_power
    write_grids = mci%write_grids
    write_all_grids = mci%write_all_grids
    read_grids = mci%read_grids
    read_grids_force = mci%read_grids_force
    initial_weights = mci%initial_weights
    apply_mappings = mci%apply_mappings
    generate_phase_space = mci%generate_phase_space
    read_phase_space = mci%read_phase_space
    overwrite_phase_space = mci%overwrite_phase_space
    phase_space_only = mci%phase_space_only
    off_shell_lines = mci%off_shell_lines
    exchange_lines = mci%exchange_lines
    extra_off_shell_lines = mci%extra_off_shell_lines
    s_channel_resonance = mci%s_channel_resonance
    single_off_shell_decays = mci%single_off_shell_decays
    double_off_shell_decays = mci%double_off_shell_decays
    single_off_shell_branchings = mci%single_off_shell_branchings
    double_off_shell_branchings = mci%double_off_shell_branchings
    massive_fsr = mci%massive_fsr
    default_jet_cut = mci%default_jet_cut
    default_mass_cut = mci%default_mass_cut
    default_energy_cut = mci%default_energy_cut
    default_q_cut = mci%default_q_cut
    n_events = mci%n_events
    n_events_warmup = mci%n_events_warmup
    unweighted = mci%unweighted
    safety_factor = mci%safety_factor
    write_events_raw = mci%write_events_raw
    write_events = mci%write_events
    write_events_format = mci%write_events_format
    read_events_raw = mci%read_events_raw
    read_events_force = mci%read_events_force
    fragment = mci%fragment
    fragmentation_method = mci%fragmentation_method
    pygive = mci%pygive
    chattiness = mci%chattiness
    screen_results = mci%screen_results
    screen_histograms = mci%screen_histograms
    screen_diagnostics = mci%screen_diagnostics
    show_pythia_banner = mci%show_pythia_banner
    write_logfile = mci%write_logfile
    show_input = mci%show_input
    show_results = mci%show_results
    show_phase_space = mci%show_phase_space
    show_cuts = mci%show_cuts
    show_histories = mci%show_histories
    show_history = mci%show_history
    show_weights = mci%show_weights
    show_event = mci%show_event
    show_histograms = mci%show_histograms
    show_overflow = mci%show_overflow
    show_excess = mci%show_excess
    plot_width = mci%plot_width
    plot_height = mci%plot_height
    read(u, nml=process_input)
    read(u, nml=integration_input)
    read(u, nml=simulation_input)
    read(u, nml=diagnostics_input)
    call read(u, mci%par)
    call read(u, mci%beam)
    mci%process_id = process_id
    mci%sqrts = sqrts
    mci%luminosity = luminosity
    mci%polarized_beams = polarized_beams
    mci%calls(:,1:3) = calls(:,1:3)
    mci%seed = seed
    mci%accuracy_goal = accuracy_goal
    mci%efficiency_goal = efficiency_goal
    mci%stratified = stratified
    mci%use_efficiency = use_efficiency
    mci%weights_power = weights_power
    mci%write_grids = write_grids
    mci%write_all_grids = write_all_grids
    mci%read_grids = read_grids
    mci%read_grids_force = read_grids_force
    mci%initial_weights = initial_weights
    mci%apply_mappings = apply_mappings
    mci%generate_phase_space = generate_phase_space
    mci%read_phase_space = read_phase_space
    mci%overwrite_phase_space = overwrite_phase_space
    mci%phase_space_only = phase_space_only
    mci%off_shell_lines = off_shell_lines
    mci%exchange_lines = exchange_lines
    mci%extra_off_shell_lines = extra_off_shell_lines
    mci%s_channel_resonance = s_channel_resonance
    mci%single_off_shell_decays = single_off_shell_decays
    mci%double_off_shell_decays = double_off_shell_decays
    mci%single_off_shell_branchings = single_off_shell_branchings
    mci%double_off_shell_branchings = double_off_shell_branchings
    mci%massive_fsr = massive_fsr
    mci%default_jet_cut = default_jet_cut
    mci%default_mass_cut = default_mass_cut
    mci%default_energy_cut = default_energy_cut
    mci%default_q_cut = default_q_cut
    mci%n_events = n_events
    mci%n_events_warmup = n_events_warmup
    mci%unweighted = unweighted
    mci%safety_factor = safety_factor
    mci%write_events_raw = write_events_raw
    mci%write_events = write_events
    mci%write_events_format = write_events_format
    mci%read_events_raw = read_events_raw
    mci%read_events_force = read_events_force
    mci%fragment = fragment
    mci%fragmentation_method = fragmentation_method
    mci%pygive = pygive
    mci%chattiness = chattiness
    mci%screen_results = screen_results
    mci%screen_histograms = screen_histograms
    mci%screen_diagnostics = screen_diagnostics
    mci%show_pythia_banner = show_pythia_banner
    mci%write_logfile = write_logfile
    mci%show_input = show_input
    mci%show_results = show_results
    mci%show_phase_space = show_phase_space
    mci%show_cuts = show_cuts
    mci%show_histories = show_histories
    mci%show_history = show_history
    mci%show_weights = show_weights
    mci%show_event = show_event
    mci%show_histograms = show_histograms
    mci%show_overflow = show_overflow
    mci%show_excess = show_excess
    mci%plot_width = plot_width
    mci%plot_height = plot_height
  end subroutine mc_input_read_unit

  subroutine mc_input_write_unit(u, mci, n_channels)
    integer, intent(in) :: u
    type(monte_carlo_input), intent(in) :: mci
    integer, intent(in) :: n_channels
    write(u,*) '&process_input'
    write(u,*) 'process_id         = ', mci%process_id
    write(u,'(1x,A,1x,G10.5)') 'sqrts              =', mci%sqrts
    write(u,'(1x,A,1x,G10.5)') 'luminosity         =', mci%luminosity
    write(u,*) 'polarized_beams    =', mci%polarized_beams
    write(u,*) '/'
    write(u,*) '&integration_input'
    write(u,*) 'calls ='
    write(u,'(3x,I3,1x,I8)') mci%calls(:,1:3)
    write(u,*) 'seed               =', mci%seed
    write(u,'(1x,A,1x,G10.5)') 'accuracy_goal      =', mci%accuracy_goal
    write(u,'(1x,A,1x,G10.5)') 'efficiency_goal    =', mci%efficiency_goal
    write(u,*) 'stratified         =', mci%stratified
    write(u,*) 'use_efficiency     =', mci%use_efficiency
    write(u,'(1x,A,1x,G10.5)') 'weights_power      =', mci%weights_power
    write(u,*) 'write_grids        =', mci%write_grids
    write(u,*) 'write_all_grids    =', mci%write_all_grids
    write(u,*) 'read_grids         =', mci%read_grids
    write(u,*) 'read_grids_force   =', mci%read_grids_force
    write(u,*) 'initial_weights    ='
    write(u,'(3x,G10.5)') mci%initial_weights(1:n_channels)
    !write(u,*) 'apply_mappings     =', mci%apply_mappings
    write(u,*) 'generate_phase_space       =', mci%generate_phase_space
    write(u,*) 'read_phase_space           =', mci%read_phase_space
    write(u,*) 'overwrite_phase_space      =', mci%overwrite_phase_space
    write(u,*) 'phase_space_only           =', mci%phase_space_only
    write(u,*) 'off_shell_lines            =', mci%off_shell_lines
    write(u,*) 'exchange_lines             =', mci%exchange_lines
    write(u,*) 'extra_off_shell_lines      =', mci%extra_off_shell_lines
    write(u,*) 's_channel_resonance        =', mci%s_channel_resonance
    write(u,*) 'single_off_shell_decays     =', mci%single_off_shell_decays
    write(u,*) 'double_off_shell_decays     =', mci%double_off_shell_decays
    write(u,*) 'single_off_shell_branchings =', mci%single_off_shell_branchings
    write(u,*) 'double_off_shell_branchings =', mci%double_off_shell_branchings
    write(u,*) 'massive_fsr                 =', mci%massive_fsr
    write(u,*) 'default_jet_cut    =', mci%default_jet_cut
    write(u,*) 'default_mass_cut   =', mci%default_mass_cut
    write(u,*) 'default_energy_cut =', mci%default_energy_cut
    write(u,*) 'default_q_cut      =', mci%default_q_cut
    write(u,*) '/'
    write(u,*) '&simulation_input'
    write(u,'(1x,A,1x,G10.5)') 'n_events             =', mci%n_events
    write(u,'(1x,A,1x,G10.5)') 'n_events_warmup      =', mci%n_events_warmup
    write(u,*) 'unweighted           =', mci%unweighted
    write(u,'(1x,A,1x,G10.5)') 'safety_factor        =', mci%safety_factor
    write(u,*) 'write_events_raw     =', mci%write_events_raw
    write(u,*) 'write_events         =', mci%write_events
    write(u,*) 'write_events_format  =', mci%write_events_format
    write(u,*) 'read_events_raw      =', mci%read_events_raw
    write(u,*) 'read_events_force    =', mci%read_events_force
    write(u,*) 'fragment             =', mci%fragment
    write(u,*) 'fragmentation_method =', mci%fragmentation_method
    if (mci%pygive == ' ') then
       write(u,*) 'pygive = " "'
    else
       write(u,*) 'pygive = "', trim(mci%pygive), '"'
    end if
    write(u,*) '/'
    write(u,*) '&diagnostics_input'
    write(u,*) 'chattiness         =', mci%chattiness
    write(u,*) 'screen_results     =', mci%screen_results
    write(u,*) 'screen_histograms  =', mci%screen_histograms
    write(u,*) 'screen_diagnostics =', mci%screen_diagnostics
    write(u,*) 'show_pythia_banner =', mci%show_pythia_banner
    write(u,*) 'write_logfile      =', mci%write_logfile
    write(u,*) 'show_input         =', mci%show_input
    write(u,*) 'show_result        =', mci%show_results
    write(u,*) 'show_phase_space   =', mci%show_phase_space
    write(u,*) 'show_cuts          =', mci%show_cuts
    write(u,*) 'show_histories     =', mci%show_histories
    write(u,*) 'show_history       =', mci%show_history
    write(u,*) 'show_weights       =', mci%show_weights
    write(u,*) 'show_event         =', mci%show_event
    write(u,*) 'show_histograms    =', mci%show_histograms
    write(u,*) 'show_overflow      =', mci%show_overflow
    write(u,*) 'show_excess        =', mci%show_excess
    write(u,*) 'plot_width         =', mci%plot_width
    write(u,*) 'plot_height        =', mci%plot_height
    write(u,*) '/'
    call write(u, mci%par)
    call write(u, mci%beam)
  end subroutine mc_input_write_unit

  subroutine mc_input_read_name(filename, mci)
    character(len=*), intent(in) :: filename
    type(monte_carlo_input), intent(inout) :: mci
    character(len=max(len(filename), len(DEFAULT_FILENAME))+3) :: file
    integer :: u
    logical :: exist
    file = trim(filename)//".in"
    inquire (file=trim(file), exist=exist)
    if (.not.exist) then
       file = DEFAULT_FILENAME//".in"
       inquire (file=trim(file), exist=exist)
       if (.not.exist)  &
            & call msg_fatal ("Input file "//trim(file)//" not found.")
    end if
    call msg_message (" Reading process data from file "//trim(file))
    u = free_unit()
    open(u, file=trim(file), action='read', status='old')
    call mc_input_read_unit(u, mci)
    close(u)
  end subroutine mc_input_read_name

  subroutine mc_input_write_name(name, mci, n_channels)
    character(len=*) :: name
    type(monte_carlo_input), intent(inout) :: mci
    integer, intent(in) :: n_channels
    integer :: u
    u = free_unit()
    open(u, file=name, action='write', status='replace')
    call mc_input_write_unit(u, mci, n_channels)
    close(u)
  end subroutine mc_input_write_name

  subroutine mc_input_broadcast(mci, root, domain, error)
    type(monte_carlo_input), intent(inout) :: mci
    integer, intent(in) :: root
    integer, intent(in), optional :: domain
    integer, intent(out), optional :: error
    call mpi90_broadcast(mci%process_id, root, domain, error)
    call mpi90_broadcast(mci%sqrts, root, domain, error)
    call mpi90_broadcast(mci%luminosity, root, domain, error)
    call mpi90_broadcast(mci%polarized_beams, root, domain, error)
    call mpi90_broadcast(mci%calls, root, domain, error)
    call mpi90_broadcast(mci%seed, root, domain, error)
    call mpi90_broadcast(mci%accuracy_goal, root, domain, error)
    call mpi90_broadcast(mci%efficiency_goal, root, domain, error)
    call mpi90_broadcast(mci%stratified, root, domain, error)
    call mpi90_broadcast(mci%use_efficiency, root, domain, error)
    call mpi90_broadcast(mci%weights_power, root, domain, error)
    call mpi90_broadcast(mci%write_grids, root, domain, error)
    call mpi90_broadcast(mci%write_all_grids, root, domain, error)
    call mpi90_broadcast(mci%read_grids, root, domain, error)
    call mpi90_broadcast(mci%read_grids_force, root, domain, error)
    call mpi90_broadcast(mci%initial_weights, root, domain, error)
    call mpi90_broadcast(mci%apply_mappings, root, domain, error)
    call mpi90_broadcast(mci%generate_phase_space, root, domain, error)
    call mpi90_broadcast(mci%read_phase_space, root, domain, error)
    call mpi90_broadcast(mci%overwrite_phase_space, root, domain, error)
    call mpi90_broadcast(mci%phase_space_only, root, domain, error)
    call mpi90_broadcast(mci%off_shell_lines, root, domain, error)
    call mpi90_broadcast(mci%exchange_lines, root, domain, error)
    call mpi90_broadcast(mci%extra_off_shell_lines, root, domain, error)
    call mpi90_broadcast(mci%s_channel_resonance, root, domain, error)
    call mpi90_broadcast(mci%single_off_shell_decays, root, domain, error)
    call mpi90_broadcast(mci%double_off_shell_decays, root, domain, error)
    call mpi90_broadcast(mci%single_off_shell_branchings, root, domain, error)
    call mpi90_broadcast(mci%double_off_shell_branchings, root, domain, error)
    call mpi90_broadcast(mci%massive_fsr, root, domain, error)
    call mpi90_broadcast(mci%default_jet_cut, root, domain, error)
    call mpi90_broadcast(mci%default_mass_cut, root, domain, error)
    call mpi90_broadcast(mci%default_energy_cut, root, domain, error)
    call mpi90_broadcast(mci%default_q_cut, root, domain, error)
    call mpi90_broadcast(mci%n_events, root, domain, error)
    call mpi90_broadcast(mci%n_events_warmup, root, domain, error)
    call mpi90_broadcast(mci%unweighted, root, domain, error)
    call mpi90_broadcast(mci%safety_factor, root, domain, error)
    call mpi90_broadcast(mci%write_events_raw, root, domain, error)
    call mpi90_broadcast(mci%write_events, root, domain, error)
    call mpi90_broadcast(mci%write_events_format, root, domain, error)
    call mpi90_broadcast(mci%read_events_raw, root, domain, error)
    call mpi90_broadcast(mci%read_events_force, root, domain, error)
    call mpi90_broadcast(mci%fragment, root, domain, error)
    call mpi90_broadcast(mci%fragmentation_method, root, domain, error)
    !call mpi90_broadcast(mci%pygive, root, domain, error)
    call mpi90_broadcast(mci%chattiness, root, domain, error)
    call mpi90_broadcast(mci%screen_results, root, domain, error)
    call mpi90_broadcast(mci%screen_histograms, root, domain, error)
    call mpi90_broadcast(mci%screen_diagnostics, root, domain, error)
    call mpi90_broadcast(mci%show_pythia_banner, root, domain, error)
    call mpi90_broadcast(mci%write_logfile, root, domain, error)
    call mpi90_broadcast(mci%show_input, root, domain, error)
    call mpi90_broadcast(mci%show_results, root, domain, error)
    call mpi90_broadcast(mci%show_phase_space, root, domain, error)
    call mpi90_broadcast(mci%show_cuts, root, domain, error)
    call mpi90_broadcast(mci%show_histories, root, domain, error)
    call mpi90_broadcast(mci%show_history, root, domain, error)
    call mpi90_broadcast(mci%show_weights, root, domain, error)
    call mpi90_broadcast(mci%show_event, root, domain, error)
    call mpi90_broadcast(mci%show_histograms, root, domain, error)
    call mpi90_broadcast(mci%show_overflow, root, domain, error)
    call mpi90_broadcast(mci%show_excess, root, domain, error)
    call mpi90_broadcast(mci%plot_width, root, domain, error)
    call mpi90_broadcast(mci%plot_height, root, domain, error)
    call mpi90_broadcast(mci%par, root, domain, error)
    call mpi90_broadcast(mci%beam, root, domain, error)
  end subroutine mc_input_broadcast
  subroutine character_broadcast(string, root, domain, error)
    character(len=*), intent(inout) :: string
    integer, intent(in) :: root
    integer, intent(in), optional :: domain
    integer, intent(out), optional :: error
    integer, dimension(:), allocatable :: buffer
    allocate(buffer(size(transfer(string, buffer))))
    buffer = transfer(string, buffer)
    call mpi90_broadcast(buffer, root, domain, error)
    string = transfer(buffer, string)
    deallocate(buffer)
  end subroutine character_broadcast

  subroutine parameter_set_broadcast(par, root, domain, error)
    type(parameter_set), intent(inout) :: par
    integer, intent(in) :: root
    integer, intent(in), optional :: domain
    integer, intent(out), optional :: error
    integer, dimension(:), allocatable :: buffer
    allocate(buffer(size(transfer(par, buffer))))
    buffer = transfer(par, buffer)
    call mpi90_broadcast(buffer, root, domain, error)
    par = transfer(buffer, par)
    deallocate(buffer)
  end subroutine parameter_set_broadcast

  subroutine set_rho_in (rho, code_in, hel_state_in, n_hel_in, pol1, pol2)
    real(kind=default), dimension(:,:), intent(inout) :: rho
    integer, dimension(:,:), intent(in) :: code_in, hel_state_in
    real(kind=default), dimension(3), intent(in) :: pol1, pol2
    integer, intent(in) :: n_hel_in
    real(kind=default), dimension(-1:1,-1:1) :: rho_tmp
    integer :: h
    call spin_density_double (rho_tmp, code_in(:,1), pol1, pol2)
    if (any(abs(hel_state_in)>1))  call msg_fatal &
         & (" Can't happen: In-state helicity greater than 1 encountered.")
    do h=1, n_hel_in
       rho(h,:) = rho_tmp(hel_state_in(1,h), hel_state_in(2,h))
    end do
  end subroutine set_rho_in

  subroutine spin_density_double (rho, code, pol1, pol2)
    real(kind=default), dimension(-1:1,-1:1), intent(inout) :: rho
    integer, dimension(2), intent(in) :: code
    real(kind=default), dimension(3), intent(in) :: pol1, pol2
    real(kind=default), dimension(-1:1) :: rho1, rho2
    integer :: i, j
    call spin_density_single (rho1, code(1), pol1)
    call spin_density_single (rho2, code(2), pol2)
    do i=-1,1
       do j=-1,1
          rho(i,j) = rho1(i) * rho2(j)
       end do
    end do
  end subroutine spin_density_double

  subroutine spin_density_single (rho, code, pol)
    real(kind=default), dimension(-1:1), intent(inout) :: rho
    integer, intent(in) :: code
    real(kind=default), dimension(3), intent(in) :: pol
    real(kind=default), dimension(3) :: pol_tmp
    integer :: mult
    mult = particle_multiplicity(code)
    pol_tmp = pol(:mult) + (1-sum(pol(:mult)))/mult
    rho = 0
    select case(mult)
    case(1)
       rho = 1
    case(2)
       rho(-1) = pol_tmp(1)
       rho(1)  = pol_tmp(2)
    case(3)
       rho(-1:1) = pol_tmp(1:3)
    end select
  end subroutine spin_density_single

  subroutine mc_state_create(mcs, mci, filename)
    character(len=*), intent(in) :: filename
    type(monte_carlo_state), intent(out) :: mcs
    type(monte_carlo_input), intent(in) :: mci
    integer :: f, i, proc_id
    call mpi90_rank (proc_id)
    mcs%process_id = mci%process_id
    mcs%sqrts = mci%sqrts
    mcs%luminosity = mci%luminosity
    mcs%polarized_beams = mci%polarized_beams
    mcs%calls(:,1:3) = mci%calls(:,1:3)
    mcs%seed = mci%seed
    mcs%accuracy_goal = mci%accuracy_goal
    mcs%efficiency_goal = mci%efficiency_goal
    mcs%stratified = mci%stratified
    mcs%use_efficiency = mci%use_efficiency
    mcs%weights_power = mci%weights_power
    mcs%write_grids = mci%write_grids
    mcs%write_all_grids = mci%write_all_grids
    mcs%read_grids = mci%read_grids
    mcs%read_grids_force = mci%read_grids_force
    mcs%initial_weights = mci%initial_weights
    mcs%apply_mappings = mci%apply_mappings
    mcs%generate_phase_space = mci%generate_phase_space
    mcs%read_phase_space = mci%read_phase_space
    mcs%overwrite_phase_space = mci%overwrite_phase_space
    mcs%phase_space_only = mci%phase_space_only
    mcs%off_shell_lines = mci%off_shell_lines
    mcs%exchange_lines = mci%exchange_lines
    mcs%extra_off_shell_lines = mci%extra_off_shell_lines
    mcs%s_channel_resonance = mci%s_channel_resonance
    mcs%single_off_shell_decays = mci%single_off_shell_decays
    mcs%double_off_shell_decays = mci%double_off_shell_decays
    mcs%single_off_shell_branchings = mci%single_off_shell_branchings
    mcs%double_off_shell_branchings = mci%double_off_shell_branchings
    mcs%massive_fsr = mci%massive_fsr
    mcs%default_jet_cut = mci%default_jet_cut
    mcs%default_mass_cut = mci%default_mass_cut
    mcs%default_energy_cut = mci%default_energy_cut
    mcs%default_q_cut = mci%default_q_cut
    mcs%n_events = mci%n_events
    mcs%n_events_warmup = mci%n_events_warmup
    mcs%unweighted = mci%unweighted
    mcs%safety_factor = mci%safety_factor
    mcs%write_events_raw = mci%write_events_raw
    mcs%write_events = mci%write_events
    mcs%write_events_format = mci%write_events_format
    mcs%read_events_raw = mci%read_events_raw
    mcs%read_events_force = mci%read_events_force
    mcs%fragment = mci%fragment
    mcs%fragmentation_method = mci%fragmentation_method
    mcs%pygive = mci%pygive
    mcs%chattiness = mci%chattiness
    mcs%screen_results = mci%screen_results
    mcs%screen_histograms = mci%screen_histograms
    mcs%screen_diagnostics = mci%screen_diagnostics
    mcs%show_pythia_banner = mci%show_pythia_banner
    mcs%write_logfile = mci%write_logfile
    mcs%show_input = mci%show_input
    mcs%show_results = mci%show_results
    mcs%show_phase_space = mci%show_phase_space
    mcs%show_cuts = mci%show_cuts
    mcs%show_histories = mci%show_histories
    mcs%show_history = mci%show_history
    mcs%show_weights = mci%show_weights
    mcs%show_event = mci%show_event
    mcs%show_histograms = mci%show_histograms
    mcs%show_overflow = mci%show_overflow
    mcs%show_excess = mci%show_excess
    mcs%plot_width = mci%plot_width
    mcs%plot_height = mci%plot_height
    mcs%par = mci%par
    call create(mcs%beam)    
    msg_level = mcs%chattiness
    mcs%n_in  = n_in(mcs%process_id)
    mcs%n_out = n_out(mcs%process_id)
    mcs%n_tot = n_tot(mcs%process_id)
    mcs%n_col = n_col(mcs%process_id)
    mcs%n_hel = n_hel(mcs%process_id)
    mcs%n_flv = n_flv(mcs%process_id)
    mcs%n_col_in = 1
    if (mcs%n_col > 0) then
       mcs%n_col_out = mcs%n_col
    else
       mcs%n_col_out = 1
    end if
    mcs%n_hel_in = n_hel_in(mcs%process_id)
    mcs%n_hel_out = n_hel_out(mcs%process_id)
    mcs%n_flv_in = n_flv_in(mcs%process_id)
    mcs%n_flv_out = n_flv_out(mcs%process_id)
    if (mcs%n_flv_in > 1) &
         & call msg_fatal ("Multiple initial states are not yet supported.")
    allocate(mcs%code(mcs%n_tot, mcs%n_flv))
    allocate(mcs%code_in(mcs%n_in, mcs%n_flv_in))
    allocate(mcs%code_out(mcs%n_out, mcs%n_flv_out))
    call flv_states(mcs%code, mcs%process_id)
    call flv_states_in(mcs%code_in, mcs%process_id)
    call flv_states_out(mcs%code_out, mcs%process_id)
    call write_process &
         & (6, mcs%process_id, mcs%code(:,1), mcs%n_in, mcs%n_out)
    do f=2, mcs%n_flv
       do i=1, mcs%n_tot
          if (particle_spin_type(mcs%code(i,f)) &
               & /= particle_spin_type(mcs%code(i,1))) call msg_fatal &
               & (" Subprocesses with different spin assignment cannot be combined.")
          if (particle_mass(mcs%code(i,f), mcs%par) &
               & /= particle_mass(mcs%code(i,1), mcs%par)) call msg_fatal &
               & (" Subprocesses with different mass assignment cannot be combined.")
       end do
    end do
    allocate(mcs%hel_state(mcs%n_tot, mcs%n_hel))
    allocate(mcs%hel_state_in(mcs%n_in, mcs%n_hel_in))
    allocate(mcs%hel_state_out(mcs%n_out, mcs%n_hel_out))
    call hel_states(mcs%hel_state, mcs%process_id)
    call hel_states_in(mcs%hel_state_in, mcs%process_id)
    call hel_states_out(mcs%hel_state_out, mcs%process_id)
    allocate(mcs%col_flow(mcs%n_tot, mcs%n_col_out))
    allocate(mcs%acl_flow(mcs%n_tot, mcs%n_col_out))
    call col_flows(mcs%col_flow, mcs%process_id)
    call acl_flows(mcs%acl_flow, mcs%process_id)
    if (mcs%n_hel == 0 .and. mcs%polarized_beams) then
       call msg_error (" Unpolarized matrix element:  Beam polarization will be ignored.")
    end if
    call set(mcs%beam, mcs%par, mcs%code(:mcs%n_in,1), mcs%sqrts, &
         &   mcs%polarized_beams, mci%beam)
    if (mcs%n_hel > 0) then
       allocate (mcs%rho_in(mcs%n_hel_in,1))
       call set_rho_in &
            & (mcs%rho_in, mcs%code_in, mcs%hel_state_in, mcs%n_hel_in, &
            &  mcs%beam(1)%polarization, mcs%beam(2)%polarization)
    else
       allocate (mcs%rho_in(1,1))
       mcs%rho_in = 1
    end if
    if (mcs%calls(1,1) <= 0) then
       mcs%calls(1,1) = 1
    end if
    if (mcs%calls(1,2) <= 0) then
       select case (mcs%n_out)
       case (:2);  mcs%calls(1,2) =  3
       case (3);   mcs%calls(1,2) =  5
       case (4);   mcs%calls(1,2) = 10
       case (5);   mcs%calls(1,2) = 15
       case (6);   mcs%calls(1,2) = 20
       case (7);   mcs%calls(1,2) = 25
       case (8:);  mcs%calls(1,2) = 30
       end select
    end if
    if (mcs%calls(1,3) <= 0) then
       mcs%calls(1,3) = 3
    end if
    if (mcs%n_events_warmup > 0) then
       mcs%calls(1,4) = 1
    end if
    if (mcs%n_events > 0 .or. mcs%luminosity > 0) then
       mcs%calls(1,5) = 1
    end if
    if (mcs%calls(2,1) <= 0) then
       select case (mcs%n_out)
       case (:2);  mcs%calls(2,1) =   5000
       case (3);   mcs%calls(2,1) =  10000
       case (4);   mcs%calls(2,1) =  20000
       case (5);   mcs%calls(2,1) =  50000
       case (6);   mcs%calls(2,1) = 100000
       case (7);   mcs%calls(2,1) = 200000
       case (8:);  mcs%calls(2,1) = 500000
       end select
    end if
    if (mcs%calls(2,2) <= 0) then
       mcs%calls(2,2) = mcs%calls(2,1)
    end if
    if (mcs%calls(2,3) <= 0) then
       mcs%calls(2,3) = mcs%calls(2,2)
    end if
    if (mcs%n_events_warmup > 0) then
       mcs%calls(2,4) = mcs%n_events_warmup
    end if
    if (mcs%n_events > 0) then
       mcs%calls(2,5) = mcs%n_events
    end if
    mcs%n_iterations_whizard = &
         & 1 + mcs%calls(1,2) + 1 + mcs%calls(1,4) + mcs%calls(1,5)
    mcs%n_iterations_vamp = sum(mcs%calls(1,:))
    if (proc_id == WHIZARD_ROOT) then
       call forest_count(trim(filename), trim(mcs%process_id), &
            & mcs%code, mcs%par, mcs%n_channels, &
            & mcs%generate_phase_space, mcs%read_phase_space, &
            & mcs%overwrite_phase_space, &
            & mcs%single_off_shell_decays, mcs%double_off_shell_decays, &
            & mcs%single_off_shell_branchings, mcs%double_off_shell_branchings, &
            & mcs%massive_fsr, &
            & mcs%off_shell_lines, mcs%exchange_lines, mcs%extra_off_shell_lines, &
            & mcs%s_channel_resonance)
    end if
    call mpi90_broadcast(mcs%n_channels, WHIZARD_ROOT)
    call create(mcs%f, mcs%n_in, mcs%n_out, mcs%n_channels)
    if (proc_id == WHIZARD_ROOT) then
       call read(trim(filename), mcs%f, trim(mcs%process_id))
    end if
    call mpi90_broadcast(mcs%f, WHIZARD_ROOT)
    call forest_set_strfun(mcs%f, sum(mcs%beam%n_strfun))
    call forest_prepare(mcs%f, mcs%code(mcs%n_in+1:,1), mcs%par)
    call process_create(trim(mcs%process_id))
    call process_reset(mcs%par, trim(mcs%process_id))
    call create(mcs%cc)
    call make_cuts &
         & (trim(filename), &
         &  mcs%process_id, mcs%cc, mcs%code(:,1), mcs%n_in, mcs%n_out, mcs%par, &
         &  mcs%default_jet_cut, mcs%default_mass_cut, mcs%default_energy_cut, &
         &  mcs%default_q_cut)
    mcs%it = 0
    mcs%it_vamp = 0
    allocate(mcs%history(mcs%n_iterations_vamp))
    allocate(mcs%histories(mcs%n_iterations_vamp, mcs%n_channels))
    call vamp_create_history(mcs%history, verbose=.false.)
    call vamp_create_history(mcs%histories, verbose=.false.)
    allocate(mcs%logtable(mcs%n_iterations_whizard))
    mcs%logtable = " "
    allocate(mcs%weights(mcs%n_iterations_whizard, mcs%n_channels))
    mcs%weights(1,:) = mcs%initial_weights(1:mcs%n_channels)
    mcs%weights(1,:) = mcs%weights(1,:) / sum(mcs%weights(1,:))
    mcs%integral = 0
    mcs%error    = 0
    mcs%chi2     = 0
    mcs%efficiency = 0
    mcs%accuracy   = 0
    call create (mcs%evt, mcs%n_in, mcs%n_out, &
         &       mcs%n_col_out, mcs%n_hel_out, mcs%n_flv_out)
    call tao_random_create(mcs%rng, mcs%seed+proc_id)
    if (mcs%write_logfile) call write(trim(filename), mcs, mci)
    if (mcs%phase_space_only) stop
  end subroutine mc_state_create

  subroutine mc_state_destroy(mcs)
    type(monte_carlo_state), intent(inout) :: mcs
    call destroy(mcs%beam)
    deallocate(mcs%code, mcs%code_in, mcs%code_out)
    deallocate(mcs%hel_state, mcs%hel_state_in, mcs%hel_state_out)
    deallocate(mcs%col_flow, mcs%acl_flow)
    deallocate (mcs%rho_in)
    call process_destroy(trim(mcs%process_id))
    call destroy(mcs%f)
    call destroy(mcs%cc)
    deallocate(mcs%history)
    deallocate(mcs%histories)
    deallocate(mcs%logtable)
    deallocate(mcs%weights)
    call destroy(mcs%evt)
  end subroutine mc_state_destroy

  subroutine mc_state_write_unit(u, mcs, mci)
    integer, intent(in) :: u
    type(monte_carlo_state), intent(in) :: mcs
    type(monte_carlo_input), intent(in) :: mci
    integer :: pass, it0, it1
    call msg_message (" WHIZARD 1.12 (Jun 20 2001)", unit=u)
    call write_process &
         & (u, mcs%process_id, mcs%code(:,1), mcs%n_in, mcs%n_out)
    call mc_write_header_unit(u, trim(mcs%process_id), mcs%checksum)
    it0 = 0
    LEVELS: do pass=1,4
       select case (pass)
       case (1,4)
          it0 = it0 + 1
          it1 = it0
       case (2)
          it0 = it0 + 1
          it1 = it0 + mcs%calls(1,2)-1
       case (3)
          it0 = it0 + mcs%calls(1,2)
          it1 = it0
       end select
       if (it0>mcs%it) exit LEVELS
       call mc_write_result &
            & (u, min(it1,mcs%it)-it0+1, mcs%logtable(it0:min(it1,mcs%it)))
       call mc_write_hline(u)
    end do LEVELS
    if (mcs%show_input) then
       call mc_write_dline(u)
       call msg_message (" Input data:", unit=u)
       call write(u, mci, mcs%n_channels)
    end if
    if (mcs%show_results) then
       call mc_write_dline(u)
       call msg_message (" Total cross section in fb:", unit=u)
       call write_integration_results (u, mcs)
    end if
    if (mcs%show_phase_space) then
       call mc_write_dline(u)
       call msg_message (" Phase space configuration:", unit=u)
       call write(u, mcs%f)
    end if
    if (mcs%show_cuts) then
       call mc_write_dline(u)
       call msg_message (" Cut configuration:", unit=u)
       if (cut_number(mcs%cc)>0) then
          call mc_write_dline(u)
          call write(u, mcs%cc)
       end if
    end if
    if (mcs%show_event) then
       call mc_write_dline(u)
       call msg_message (" Event display:", unit=u)
       call mc_write_dline(u)
       call write(u, mcs%evt)
    end if
    if (mcs%actual_integration_done) then
       if (mcs%show_histories) then
          call mc_write_dline(u)
          call msg_message (" Channel histories [vamp]:", unit=u)
          call vamp_write_history(u, mcs%histories)
       end if
       if (mcs%show_weights) then
          call mc_write_dline(u)
          call msg_message (" Weight history [%]:", unit=u)
          call mc_write_weights_unit(u,mcs)
       end if
       if (mcs%show_history) then
          call mc_write_dline(u)
          call msg_message (" Global history [vamp]:", unit=u)
          call vamp_write_history(u, mcs%history)
       end if
    else
       call msg_message (" History information not shown (results read from file).", unit=u)
    end if
  end subroutine mc_state_write_unit

  subroutine mc_state_write_name(filename, mcs, mci)
    character(len=*), intent(in) :: filename
    type(monte_carlo_state), intent(in) :: mcs
    type(monte_carlo_input), intent(in) :: mci
    integer :: u
    u = free_unit()
    open (unit=u, action='write', status='replace', file=filename//".out")
    call mc_state_write_unit(u, mcs, mci)
    close (unit=u)
  end subroutine mc_state_write_name

  subroutine mc_write_weights_unit(u,mcs)
    integer, intent(in) :: u
    type(monte_carlo_state), intent(in) :: mcs
    integer :: it, ch
    write(u, '(1x,A6)', advance='no') '! ch |'
    do it=1, mcs%it
       write(u, '(I3)', advance='no') it
    end do
    write(u,*)
    call mc_write_hline(u)
    do ch=1, mcs%f%n_channels
       write(u, '(1x,I4,1x,A1)', advance='no') ch, '|'
       do it=1, mcs%it
          if (mcs%weights(it,ch)==0) exit
          write(u, '(I3)', advance='no') nint(100*mcs%weights(it,ch))
       end do
       write(u,*)
    end do
  end subroutine mc_write_weights_unit

  function mc_checksum(mcs, level) result(s)
    type(monte_carlo_state), intent(in) :: mcs
    integer, intent(in) :: level
    integer :: s
    integer, dimension(1) :: mold
    s = 0
    s = s + sum(transfer(mcs%process_id, mold))
    s = s + sum(transfer(mcs%par, mold))
    s = s + sum(transfer(mcs%cc, mold))
    s = s + sum(transfer(mcs%beam, mold))
!    s = s + sum(transfer(mcs%calls(:,1:level), mold))
    s = s + sum(transfer(mcs%stratified, mold))*13
    s = s + sum(transfer(mcs%polarized_beams, mold))*17
    s = s + sum(transfer(mcs%sqrts, mold))
    s = s + sum(transfer(mcs%weights_power, mold))
    if (level>4) s = s + sum(transfer(mcs%unweighted, mold))*11
    if (level>4) s = s + sum(transfer(mcs%luminosity, mold))
  end function mc_checksum
  subroutine mc_checksum_error
    call msg_fatal (" Checksum mismatch.  This file can't be used, sorry.")
  end subroutine mc_checksum_error

  subroutine mc_write_header_unit(u, process_id, checksum)
    integer, intent(in) :: u
    character(len=*), intent(in), optional :: process_id
    integer, intent(in), optional :: checksum
    if (present(process_id) .and. present(checksum)) then
       call msg_message (" ", unit=u)
       write (msg_buffer, "(A,1x,I12,1x,A)") &
            & " WHIZARD run for process "//process_id//":"// &
            & "             ( checksum = ", checksum, ")"
       call msg_message (unit=u)
    end if
    call mc_write_hline_unit(u)
    if (u /= 6 .or. msg_level > 3) call msg_message &
         & (" It      Calls   Integral[fb]  Error[fb]  Err[%]    Acc  Eff[%]   Chi2 N[It]", unit=u)
    call mc_write_hline_unit(u)
  end subroutine mc_write_header_unit
  subroutine mc_write_hline_unit(u)
    integer, intent(in) :: u
    if (u /= 6 .or. msg_level > 3)  call msg_message(repeat('-',77), unit=u)
  end subroutine mc_write_hline_unit
  subroutine mc_write_dline_unit(u)
    integer, intent(in) :: u
    if (u /= 6 .or. msg_level > 3)  call msg_message(repeat('=',77), unit=u)
  end subroutine mc_write_dline_unit
  subroutine mc_write_result_string &
       & (s, it, calls, integral, error, efficiency, accuracy, &
       &  use_efficiency, improved, chi2, iterations)
    character(len=*), intent(out) :: s
    integer, intent(in) :: it, calls
    real(kind=default), intent(in) :: integral, error, efficiency, accuracy
    logical, intent(in), optional :: use_efficiency, improved
    real(kind=default), intent(in), optional :: chi2
    integer, intent(in), optional :: iterations
    character(len=len(s)) :: buffer
    integer :: pos
    real(kind=default) :: rel_error
    s = " "; buffer = " "; pos = 1
    if (integral/=0) then
       rel_error = error/integral
    else
       rel_error = 0
    end if
    write(buffer,'(1x,I3,1x,I10,1x,1PE14.7,1x,1PE9.2,1x,2PF7.2)') &
         & it, calls, integral, error, rel_error
    call append(len_trim(buffer))
    write(buffer,'(1x,0PF7.2)') accuracy
    call append(len_trim(buffer))
    if (present(use_efficiency)) then
       if (.not.use_efficiency.and.present(improved)) then
          if (improved) then
             buffer = "*"
          end if
       end if
    end if
    call append(1)
    write(buffer,'(2PF6.2)') efficiency
    call append(len_trim(buffer))
    if (present(use_efficiency)) then
       if (use_efficiency.and.present(improved)) then
          if (improved) then
             buffer = "*"
          end if
       end if
    end if
    call append(1)
    if (present(chi2)) then
       write(buffer,'(0PF7.2)') chi2
       call append(len_trim(buffer))
       if (present(iterations)) then
          call append(1)
          write(buffer,'(I3)') iterations
          call append(len_trim(buffer))
       end if
    end if
  contains
    subroutine append(p)
      integer, intent(in) :: p
      s(pos+1:pos+p) = buffer
      pos = pos+p
      buffer = " "
    end subroutine append
  end subroutine mc_write_result_string

  subroutine mc_write_result_unit &
       & (u, it, calls, integral, error, efficiency, accuracy, &
       &  use_efficiency, improved, chi2, iterations)
    integer, intent(in) :: u
    integer, intent(in) :: it, calls
    real(kind=default), intent(in) :: integral, error, efficiency, accuracy
    logical, intent(in), optional :: use_efficiency, improved
    real(kind=default), intent(in), optional :: chi2
    integer, intent(in), optional :: iterations
    character(len=LOGTABLE_LINE_LENGTH) :: s
    call mc_write_result_string(s, it, calls, integral, error, &
         & efficiency, accuracy, use_efficiency, improved, chi2, iterations)
    call msg_result (s, u)
  end subroutine mc_write_result_unit

  subroutine mc_write_logtable_unit (u, it, s)
    integer, intent(in) :: u, it
    character(len=*), dimension(:), intent(in) :: s
    integer :: i
    do i=1, it
       call msg_result (s(i), u)
    end do
  end subroutine mc_write_logtable_unit

  subroutine write_integration_results (u, mcs)
    integer, intent(in) :: u
    type(monte_carlo_state), intent(in) :: mcs
    write (u, *) "&integration_results"
    write (u, *) "integral = ", mcs%integral
    write (u, *) "error    = ", mcs%error
    write (u, *) "chi2     = ", mcs%chi2
    write (u, *) "calls_total = ", mcs%calls(1,3) * mcs%calls(2,3)
    write (u, *) "iterations  = ", mcs%calls(1,3)
    write (u, *) "/"
  end subroutine write_integration_results
  subroutine mc_read_logtable_unit (u, it, s)
    integer, intent(in) :: u, it
    character(len=*), dimension(:), intent(out) :: s
    integer :: i, n_it, n_ignore
    if (it <= size(s)) then
       n_it = it
       n_ignore = 0
    else
       n_it = size(s)
       n_ignore = it - size(s)
    end if
    do i=1, n_it
       read(u, '(A)') s(i)
    end do
    do i=1, n_ignore
       read(u, *)
    end do
  end subroutine mc_read_logtable_unit

  function mc_get_efficiency (g) result(efficiency)
    type(vamp_grids), intent(in) :: g
    real(kind=default) :: efficiency
    real(kind=default) :: wsum
    real(kind=default), dimension(size(g%grids)) :: weights, tmp
    where(g%num_calls >= 2 .and. g%grids%f_max>0)
       weights = g%weights * g%grids%f_max
    elsewhere
       weights = 0
    end where
    wsum = sum(weights)
    if (wsum/=0) then
       efficiency = dot_product(mc_get_efficiency_vector (g), weights) / wsum
    else
       efficiency = 1
    end if
  end function mc_get_efficiency

  function mc_get_efficiency_vector (g) result(efficiency)
    type(vamp_grids), intent(in) :: g
    real(kind=default), dimension(size(g%grids)) :: efficiency
    where(g%num_calls >= 2 .and. g%grids%f_max>0)
       efficiency = g%grids%mu(1) / g%grids%f_max
    elsewhere
       efficiency = 0
    end where
  end function mc_get_efficiency_vector

  function mc_get_accuracy (integral, error, calls) result(a)
    real(kind=default), intent(in) :: integral, error
    integer, intent(in) :: calls
    real(kind=default) :: a
    if (integral/=0) then
       a = error / integral * sqrt(real(calls, kind=default))
    else
       a = 0
    end if
  end function mc_get_accuracy


end module monte_carlo_states
